use std::collections::HashMap;
use std::future::Future;

use bitfinex::positions::Position;
use tokio::task::JoinHandle;

use crate::BoxError;
use crate::pairs::PairStatus;

#[derive(Copy, Clone)]
pub enum SignalKind {
    ClosePosition,
    OpenPosition,
}

#[derive(Copy, Clone)]
pub struct EventMetadata {
    position_id: Option<u64>,
    order_id: Option<u64>,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub enum EventKind {
    NewMinimum,
    NewMaximum,
    ReachedLoss,
    ReachedBreakEven,
    ReachedMinProfit,
    ReachedGoodProfit,
    ReachedMaxLoss,
    TrailingStopSet,
    TrailingStopMoved,
    OrderSubmitted,
    NewTick,
}

#[derive(Copy, Clone)]
pub struct Event {
    kind: EventKind,
    tick: u64,
    metadata: Option<EventMetadata>,
}

impl Event {
    pub fn new(kind: EventKind, tick: u64, metadata: Option<EventMetadata>) -> Self {
        Event {
            kind,
            tick,
            metadata,
        }
    }

    fn has_metadata(&self) -> bool {
        self.metadata.is_some()
    }
}

pub struct EventDispatcher {
    event_handlers: HashMap<EventKind, Vec<Box<dyn FnMut(&Position, &PairStatus) -> JoinHandle<()>>>>,
}

impl EventDispatcher {
    pub fn new() -> Self {
        EventDispatcher {
            event_handlers: HashMap::new(),
        }
    }

    pub fn call_handler(&mut self, event: &EventKind, position: &Position, status: &PairStatus) {
        if let Some(callbacks) = self.event_handlers.get_mut(event) {
            for f in callbacks {
                f(position, status);
            }
        }
    }

    pub fn register_event_handler<F: 'static, Fut: 'static>(&mut self, event: EventKind, mut f: F)
        where
            F: FnMut(&Position, &PairStatus) -> Fut,
            Fut: Future<Output=()> + Send, {
        self.event_handlers
            .entry(event)
            .or_default()
            .push(Box::new(move |p, s| tokio::spawn(f(p, s))));
    }
}
